/*
Copyright (C) Christopher Fogelberg, Mengjie Zhang 2005

This file is part of VUWLGP.

VUWLGP is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

VUWLGP is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VUWLGP; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
  
#include <iostream>
#include <iomanip>

#include "Config.h"
#include "Instruction.h"

#include "IInstructionArgument.h"
#include "InstructionArgumentConstant.h"
#include "InstructionArgumentFeature.h"
#include "InstructionArgumentRegister.h"

#include "IInstructionOperation.h"
#include "IInstructionOperationConditional.h"
#include "InstructionOperationIflt.h"
#include "InstructionOperationPlus.h"

#include "MultiClassFitnessCase.h"

int main(int argc, char** argv) {
  // Configure the Config object:
  VUWLGP::Config<double>* config = new VUWLGP::Config<double>();
  config->numFeatures = 1;
  config->numRegisters = 10;

  config->argumentGenerators->AddElement(VUWLGP::InstructionArgumentConstant<double>::Generate);
  config->argumentGenerators->AddElement(VUWLGP::InstructionArgumentFeature<double>::Generate);
  config->argumentGenerators->AddElement(VUWLGP::InstructionArgumentRegister<double>::Generate);

  config->instructionOperations->AddElement(VUWLGP::InstructionOperationPlus<double>::Generate);
  config->instructionOperations->AddElement(VUWLGP::InstructionOperationIflt<double>::Generate);

  // Create random instructions, display them and execute them in some fitness environment.
  VUWLGP::FitnessEnvironment<double> fe(config);
  fe.WriteRegister(0, 100);
  fe.WriteRegister(1, 101);
  fe.WriteRegister(2, 102);
  fe.WriteRegister(3, 103);
  fe.WriteRegister(4, 104);
  fe.WriteRegister(5, 105);
  fe.WriteRegister(6, 106);
  fe.WriteRegister(7, 107);
  fe.WriteRegister(8, 108);
  fe.WriteRegister(9, 109);
  fe.AddCase(new VUWLGP::MultiClassFitnessCase<double>("name x y 1 fun 1.5"));

  std::cout << "Creating 5 random instructions and applying them to registers, printing them.\n"
	    << "NB: All instructions are executed, but their true/false nature is printed."
	    << std::endl;
  
  std::cout << "Starting FitnessEnvironment:" << std::endl;
  std::cout << fe.ToString() << std::endl;


  VUWLGP::Instruction<double>* instruction;
  for(unsigned int i = 0; i < 5; ++i) {
    instruction = new VUWLGP::Instruction<double>(config);
    std::cout << "\tInstruction:\t" << instruction->ToString();
    std::cout << "\t1 = true conditional or any other instruction: " << instruction->Execute(fe)
	      << std::endl;
    delete instruction;
  }

  std::cout << "Resulting FitnessEnvironment:" << std::endl;
  std::cout << fe.ToString() << std::endl << std::endl;

  // Now test the mutation function:
  std::cout << "Testing the mutate functionality: " << std::endl;
  for(unsigned int i = 0; i < 5; ++i) {
    instruction = new VUWLGP::Instruction<double>(config);

    std::cout << "Instruction: " << std::setw(27) << instruction->ToString();
    instruction->Mutate();
    std::cout << "\t\tMutant: " << std::setw(27) << instruction->ToString() << std::endl;

    delete instruction;
  }

  // Now test the Clone function a few times:
  instruction = new VUWLGP::Instruction<double>(config);
  VUWLGP::Instruction<double>* clone;

  std::cout << "\nNow testing Instruction::Clone function..." << std::endl;
  std::cout << "Instruction: " << std::setw(27) << instruction->ToString() << std::endl;
  for(unsigned int i = 0; i < 5; ++i) {
    clone = instruction->Clone();
    std::cout << "\tClone: " << std::setw(27) << clone->ToString() << std::endl;
  }

  delete config;

  return 0;
}
