/*
Copyright (C) Christopher Fogelberg, Mengjie Zhang 2005

This file is part of VUWLGP.

VUWLGP is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

VUWLGP is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VUWLGP; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
  
#include <iostream>

#include "Config.h"
#include "GenerationalTournamentPopulation.h"

#include "InstructionArgumentConstant.h"
#include "InstructionArgumentFeature.h"
#include "InstructionArgumentRegister.h"

#include "InstructionOperationPlus.h"
#include "InstructionOperationIflt.h"

#include "MultiClassProgram.h"

using namespace VUWLGP; // else the lines of code get stupidly and clumsily long

// Initialise the static members which specify the proportion selected through elitism and
// the size of the selection tournaments for other evolutionary operators:
double GenerationalTournamentPopulation<MultiClassProgram<double>, double>::proportionElitism 
  = 0.25;
unsigned int GenerationalTournamentPopulation<MultiClassProgram<double>, double>::tournamentSize
  = 3;

int main(int argc, char** argv) {
  std::cout << "Testing GenerationalTournamentPopulation and IPopulation" << std::endl;
  std::cout << "Generating a population of size 10 in size range of [2,4]." << std::endl;

  // Building the relevant config object - specify it all, even though not all of it is needed:
  Config<double>* c = new Config<double>();
  c->numRegisters = 4;
  c->numFeatures = 4;

  c->argumentGenerators->AddElement(InstructionArgumentConstant<double>::Generate);
  c->argumentGenerators->AddElement(InstructionArgumentFeature<double>::Generate);
  c->argumentGenerators->AddElement(InstructionArgumentRegister<double>::Generate);

  c->instructionOperations->AddElement(InstructionOperationPlus<double>::Generate);
  c->instructionOperations->AddElement(InstructionOperationIflt<double>::Generate);

  c->initialMinLength = 2;
  c->initialMaxLength = 4;
  c->maxLength = 4;
  c->populationSize = 10;

  c->maxGenerations = 50;

  // Building the population object, randomising the fitness of it all:
  IPopulation<MultiClassProgram<double>, double>* pop = 
    new GenerationalTournamentPopulation<MultiClassProgram<double>, double>(c);
  pop->RandomiseFitness();

  // Test that we truncate and get the floor number when doing elite:
  std::cout << "Proportion-elite = 0.25, population-size = 10 -> should be two individuals\n"
	    << "selected through elitism. Number actually selected: "
	    << static_cast<unsigned int>(0.25 * 10) << "\n\n" << std::endl;

  // Test the sorting:
  std::cout << "Programs unsorted, in order generated:\n" << pop->ToString() << "\n\n" 
	    << std::endl;

  pop->SortFittestFirst();

  std::cout << "Programs in ascending order of fitness, fittest first:\n" << pop->ToString()
	    << "\n\n" << std::endl;

  std::cout << "The Fittest Program Is:\n" << pop->GetFittestProgram()->ToString() << "\n\n"
	    << std::endl;

  // Test the tournament selection methods independently of the iterate method:
  std::cout << "Now tournament-selecting two programs independently of each other by fitness:\n"
	    << "First:\n" << pop->SelectProgByFitness()->ToString() << "\n\n"
	    << "Second:\n" << pop->SelectProgByFitness()->ToString() << "\n\n" << std::endl;

  // Test the population iteration:
  pop->IteratePopulation();
  std::cout << "Result of applying the IPopulation::IteratePopulation method:\n" 
  << pop->ToString() << "\n\n" << std::endl;

  // Test each evolutionary method individually:
  delete pop;
  c->populationSize = 1;
  c->initialMinLength = 4;
  c->initialMaxLength = 4;
  pop = new GenerationalTournamentPopulation<MultiClassProgram<double>, double>(c);
  IProgram<double>* offspring;

  std::cout << "Now testing each evolutionary operator individually on a population of one:\n"
	    << pop->ToString() << "\n" << std::endl;

  offspring = pop->MacroMutation();
  std::cout << "MacroMutation:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  offspring = pop->MicroMutation();
  std::cout << "MicroMutation:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  offspring = pop->CrossoverFree();
  std::cout << "CrossoverFree:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  offspring = pop->CrossoverGA();
  std::cout << "CrossoverGA:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  offspring = pop->CrossoverHomologous();
  std::cout << "CrossoverHomologous:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  offspring = pop->AddRandomInstruction();
  std::cout << "AddRandomInstruction:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  offspring = pop->RemoveRandomInstruction();
  std::cout << "RemoveRandomInstruction:\n" << offspring->ToString() << "\n" << std::endl;
  delete offspring;

  delete pop;
  delete c;

  return 0;
}
